#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import division, unicode_literals

import json
import logging
import re
from uuid import uuid4

from . import neo4j
from .util import deprecated
from .xmlutil import xml_to_geoff

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


logger = logging.getLogger(__name__)

SIMPLE_NAME = re.compile(r"^[A-Za-z_][0-9A-Za-z_]*$")


class ConstraintViolation(ValueError):
    """ Raised on an attempt to merge a unique path onto a non-unique path.
    """
    pass


class Subgraph(object):

    @classmethod
    def load(cls, file):
        """ Load Geoff data from a file into a Subgraph.
        """
        return cls(file.read())

    @classmethod
    def load_xml(cls, file, prefixes=None):
        """ Load and convert data within an XML file into a Subgraph.
        """
        src = file.read().encode("utf-8")
        file.close()
        return cls(xml_to_geoff(src, prefixes=prefixes))

    def __init__(self, source):
        """ Create a subgraph from Geoff source.
        """
        self._source = source
        parser = _Parser(self._source)
        self._nodes, self._rels, self._index_entries = parser.parse()

    def save(self, file):
        """ Save subgraph to a file in Geoff format.
        """
        pass  #TODO: actually write this bit

    @property
    def source(self):
        """ Return the Geoff source for this Subgraph.
        """
        return self._source

    @property
    def nodes(self):
        """ Return all nodes within this Subgraph.
        """
        return dict(self._nodes)

    @property
    def relationships(self):
        """ Return all relationships within this Subgraph.
        """
        return list(self._rels)

    @property
    def index_entries(self):
        """ Return all index entries within this Subgraph.
        """
        return dict(self._index_entries)

    @property
    def _indexed_nodes(self):
        """ Return the set of nodes which have one or more index entries
        pointing to them.
        """
        return dict(
            (entry.node.name, entry.node)
            for entry in self._index_entries.values()
        )

    @property
    def _related_nodes(self):
        """ Return the set of nodes which are involved in at least one
        relationship.
        """
        nodes = dict(
            (rel.start_node.name, rel.start_node)
            for rel in self._rels
        )
        nodes.update(dict(
            (rel.end_node.name, rel.end_node)
            for rel in self._rels
        ))
        return nodes

    @property
    def _odd_nodes(self):
        """ Return the set of nodes which have no index entries pointing to
        them and are involved in no relationships.
        """
        return dict(
            (name, node)
            for name, node in self._nodes.items()
            if name not in self._indexed_nodes
            if name not in self._related_nodes
        )

    def _get_relationship_query(self, unique):
        # will only work in 1.8.1 and above
        start, create, set, output_names, params = [], [], [], [], {}
        # determine query inputs
        input_names, input_entries = list(self._indexed_nodes.keys()), []
        for name in input_names:
            entry = [
                entry
                for entry in self._index_entries.values()
                if entry.node.name == name
            ][0]
            input_entries.append(entry)
        #
        def node_pattern(i, name):
            if name in input_names:
                return "(in{0})".format(input_names.index(name))
            elif name in output_names:
                return "(out{0})".format(output_names.index(name))
            else:
                try:
                    params["sp{0}".format(i)] = self._nodes[name].properties
                except KeyError:
                    raise SystemError("Broken internal reference - "
                                      "node '{0}' not found".format(name))
                output_names.append(name)
                return "(out{0} {{{1}}})".format(output_names.index(name), "sp{0}".format(i))
            #
        def rel_pattern(i, type_, properties):
            if properties:
                params["rp{0}".format(i)] = properties
                return "-[:`{0}` {{{1}}}]->".format(type_, "rp{0}".format(i))
            else:
                return "-[:`{0}`]->".format(type_)
            #
        # build start clause from inputs
        for i, entry in enumerate(input_entries):
            start.append("in{0} = node:{1}(`{2}`={{val{0}}})".format(i, entry.index_name, entry.key))
            params["val{0}".format(i)] = entry.value
            set.append("SET in{0} = {{pa{0}}}".format(i))
            params["pa{0}".format(i)] = entry.node.properties
        for i, rel in enumerate(self._rels):
            # build and append pattern
            create.append(node_pattern(2 * i, rel.start_node.name) + \
                          rel_pattern(i, rel.type, rel.properties) + \
                          node_pattern(2 * i + 1, rel.end_node.name)
            )
        if start:
            query = "START {0}\n{1} {2}".format(
                ",\n      ".join(start),
                "CREATE UNIQUE" if unique else "CREATE",
                ",\n              ".join(create),
            )
        else:
            query = "CREATE {0}".format(
                ",\n              ".join(create),
            )
        if set:
            query += "".join("\n" + line for line in set)
        if output_names:
            query += "\nRETURN {0}".format(
                ",\n       ".join("out{0}".format(i) for i in range(len(output_names))),
            )
        return query, params, output_names

    def _execute_load_batch(self, graph_db, unique):
        # build batch request
        batch = neo4j.WriteBatch(graph_db)
        # 1. indexed nodes
        index_entries = list(self.index_entries.values())
        for entry in index_entries:
            batch.get_or_create_indexed_node(entry.index_name, entry.key, entry.value, entry.node.properties)
        # 2. related nodes
        if self._rels:
            query, params, related_names = self._get_relationship_query(unique)
            batch.append_cypher(query, params)
        else:
            related_names = []
        # 3. odd nodes
        odd_names = list(self._odd_nodes.keys())
        for name in odd_names:
            batch.create(self._nodes[name].properties)
        # submit batch unless empty (in which case bail out and return nothing)
        if batch:
            responses = batch.submit()
        else:
            return {}
        # parse response and build return value
        return_nodes = {}
        # 1. unique_nodes
        for i, entry in enumerate(index_entries):
            return_nodes[entry.node.name] = responses.pop(0)
        # 2. related nodes
        if self._rels:
            r = responses.pop(0)
            if r:
                for i, value in enumerate(r):
                    return_nodes[related_names[i]] = value
        # 3. odd nodes
        for i, name in enumerate(odd_names):
            return_nodes[name] = responses.pop(0)
        return return_nodes

    @deprecated("The geoff module is deprecated, "
                "use the load2neo server extension instead")
    def insert_into(self, graph_db):
        """ Insert subgraph into graph database using Cypher CREATE.
        """
        return self._execute_load_batch(graph_db, False)

    @deprecated("The geoff module is deprecated, "
                "use the load2neo server extension instead")
    def merge_into(self, graph_db):
        """ Merge subgraph into graph database using Cypher CREATE UNIQUE.
        """
        try:
            return self._execute_load_batch(graph_db, True)
        except neo4j.BatchError as err:
            if err.__class__.__name__ == "UniquePathNotUniqueException":
                err = ConstraintViolation(
                    "Unable to merge relationship onto multiple "
                    "existing relationships."
                )
            raise err


class _Parser(object):

    JSON_BOOLEAN = re.compile(r'(true|false)', re.IGNORECASE)
    JSON_NUMBER  = re.compile(r'(-?(0|[1-9]\d*)(\.\d+)?(e[+-]?\d+)?)', re.IGNORECASE)
    JSON_STRING  = re.compile(r'''(".*?(?<!\\)")''')
    NAME         = re.compile(r'(\w+)', re.UNICODE)
    WHITESPACE   = re.compile(r'(\s*)')

    def __init__(self, source):
        self.source = source       # original source data
        self.n = 0                 # pointer to next position to be parsed

    def _unexpected_character(self):
        next_char = self.peek()
        message = "Unexpected character {0} at position {1}".format(repr(next_char), self.n)
        return SyntaxError(message)

    def peek(self, length=1):
        return self.source[self.n:self.n + length]

    def parse(self):
        # FIRST PASS - work through source, extracting recognised elements
        self.n = 0
        elements = []   # temporary store for extracted elements
        self.parse_pattern(self.WHITESPACE)
        while self.n < len(self.source):
            element = self.parse_element()
            if isinstance(element, dict):
                if elements:
                    elements[-1].properties = element
                else:
                    raise TypeError("Property map cannot occur as first element.")
            elif isinstance(element, tuple):
                elements.extend(element)
            elif element is not None:
                elements.append(element)
            self.parse_pattern(self.WHITESPACE)
        # SECOND PASS - consolidate into collections of similar elements
        nodes, rels, index_entries = {}, [], {}
        def set_node(name, node):
            if name is None:
                nodes[uuid4()] = node
            elif name in nodes:
                nodes[name].properties.update(node.properties)
            else:
                nodes[name] = node
        for element in elements:
            if isinstance(element, AbstractNode):
                set_node(element.name, element)
            elif isinstance(element, AbstractRelationship):
                set_node(element.start_node.name, element.start_node)
                set_node(element.end_node.name, element.end_node)
                element.start_node = nodes[element.start_node.name]
                element.end_node = nodes[element.end_node.name]
                rels.append(element)
            elif isinstance(element, AbstractIndexEntry):
                set_node(element.node.name, element.node)
                element.node = nodes[element.node.name]
                key = (element.index_name, element.key, element.value, element.node.name)
                index_entries[key] = element
            else:
                raise TypeError("Unexpected element type '{0}'".format(element.__class__.__name__))
        return nodes, rels, index_entries

    def parse_array(self):
        items = []
        self.parse_literal("[")
        self.parse_pattern(self.WHITESPACE)
        next_char = self.peek()
        if next_char != "]":
            if next_char == '"':
                value_pattern = self.JSON_STRING
            elif next_char in "-0123456789":
                value_pattern = self.JSON_NUMBER
            elif next_char in "tf":
                value_pattern = self.JSON_BOOLEAN
            else:
                raise self._unexpected_character()
            items.append(self.parse_pattern(value_pattern, json.loads))
            self.parse_pattern(self.WHITESPACE)
            while self.peek() == ",":
                self.parse_literal(",")
                self.parse_pattern(self.WHITESPACE)
                items.append(self.parse_pattern(value_pattern, json.loads))
                self.parse_pattern(self.WHITESPACE)
        self.parse_literal("]")
        return items

    def parse_comment(self):
        self.parse_literal("/*")
        m = self.n
        while self.n < len(self.source) and self.peek(2) != "*/":
            self.n += 1
        comment = self.source[m:self.n]
        self.n += 2
        return comment

    def parse_element(self):
        element = None
        next_char = self.peek()
        if next_char == "(":
            node = self.parse_node()
            if self.peek(2) == "<=":
                # index entry
                self.parse_literal("<=")
                index_name, key, value = self.parse_index_point()
                element = AbstractIndexEntry(index_name, key, value, node)
            else:
                # path
                rels = []
                while self.n < len(self.source) and not self.parse_pattern(self.WHITESPACE):
                    next_char = self.peek()
                    if next_char == "-":
                        rel = self.parse_forward_path(node)
                        node = rel.end_node
                        rels.append(rel)
                    elif next_char == "<":
                        rel = self.parse_reverse_path(node)
                        node = rel.start_node
                        rels.append(rel)
                    else:
                        raise self._unexpected_character()
                if rels:
                    element = tuple(rels)
                else:
                    element = node
        elif next_char == "|":
            index_name, key, value = self.parse_index_point()
            self.parse_literal("=>")
            node = self.parse_node()
            element = AbstractIndexEntry(index_name, key, value, node)
        elif next_char == "/":
            self.parse_comment()
        elif next_char == "{":
            element = self.parse_property_map()
        else:
            raise self._unexpected_character()
        return element

    def parse_forward_path(self, start_node):
        self.parse_literal("-")
        rel = self.parse_relationship(start_node=start_node)
        self.parse_literal("->")
        rel.start_node = start_node
        rel.end_node = self.parse_node()
        return rel

    def parse_index_point(self):
        self.parse_literal("|")
        self.parse_pattern(self.WHITESPACE)
        index_name = self.parse_name()
        self.parse_pattern(self.WHITESPACE)
        next_char = self.peek()
        if next_char == "{":
            key, value = self.parse_property_pair()
            self.parse_pattern(self.WHITESPACE)
            self.parse_literal("|")
        elif next_char == "|":
            self.parse_literal("|")
            self.parse_pattern(self.WHITESPACE)
            key, value = self.parse_property_pair()
        else:
            raise self._unexpected_character()
        return index_name, key, value

    def parse_key_value_pair(self):
        key = self.parse_name()
        self.parse_pattern(self.WHITESPACE)
        self.parse_literal(":")
        self.parse_pattern(self.WHITESPACE)
        value = self.parse_value()
        return key, value

    def parse_literal(self, literal):
        len_literal = len(literal)
        test = self.source[self.n:self.n + len_literal]
        if test == literal:
            self.n += len_literal
            return literal
        else:
            raise self._unexpected_character()

    def parse_name(self):
        if self.peek() == "\"":
            return self.parse_pattern(self.JSON_STRING, json.loads)
        else:
            return self.parse_pattern(self.NAME)

    def parse_node(self):
        self.parse_literal("(")
        self.parse_pattern(self.WHITESPACE)
        next_char = self.peek()
        if next_char == ")":
            node = AbstractNode(None, None)
        elif next_char == "{":
            node = AbstractNode(None, self.parse_property_map())
        else:
            name = self.parse_name()
            if not self.parse_pattern(self.WHITESPACE):
                node = AbstractNode(name, None)
            else:
                node = AbstractNode(name, self.parse_property_map())
        self.parse_pattern(self.WHITESPACE)
        self.parse_literal(")")
        return node

    def parse_pattern(self, pattern, decoder=None):
        m = pattern.match(self.source, self.n)
        if m:
            self.n += len(m.group(0))
            value= m.group(0)
        else:
            raise self._unexpected_character()
        if decoder:
            value = decoder(value)
        return value

    def parse_property_map(self):
        properties = []
        self.parse_literal("{")
        self.parse_pattern(self.WHITESPACE)
        if self.peek() != "}":
            properties.append(self.parse_key_value_pair())
            self.parse_pattern(self.WHITESPACE)
            while self.peek() == ",":
                self.parse_literal(",")
                self.parse_pattern(self.WHITESPACE)
                properties.append(self.parse_key_value_pair())
                self.parse_pattern(self.WHITESPACE)
        self.parse_literal("}")
        return dict(properties)

    def parse_property_pair(self):
        self.parse_literal("{")
        self.parse_pattern(self.WHITESPACE)
        key, value = self.parse_key_value_pair()
        self.parse_pattern(self.WHITESPACE)
        self.parse_literal("}")
        return key, value

    def parse_relationship(self, start_node=None, end_node=None):
        self.parse_literal("[")
        self.parse_pattern(self.WHITESPACE)
        next_char = self.peek()
        if next_char != ":":
            # read and ignore relationship name, if present
            self.parse_name()
            self.parse_pattern(self.WHITESPACE)
        self.parse_literal(":")
        type = self.parse_name()
        self.parse_pattern(self.WHITESPACE)
        next_char = self.peek()
        if next_char == "{":
            rel = AbstractRelationship(start_node, type, self.parse_property_map(), end_node)
            self.parse_pattern(self.WHITESPACE)
        else:
            rel = AbstractRelationship(start_node, type, None, end_node)
        self.parse_literal("]")
        return rel

    def parse_reverse_path(self, end_node):
        self.parse_literal("<-")
        rel = self.parse_relationship(end_node=end_node)
        self.parse_literal("-")
        rel.start_node = self.parse_node()
        rel.end_node = end_node
        return rel

    def parse_value(self):
        next_char = self.peek()
        if next_char == "[":
            value = self.parse_array()
        elif next_char == '"':
            value = self.parse_pattern(self.JSON_STRING, json.loads)
        elif next_char in "-0123456789":
            value = self.parse_pattern(self.JSON_NUMBER, json.loads)
        elif next_char == "t":
            self.parse_literal("true")
            value = True
        elif next_char == "f":
            self.parse_literal("false")
            value = False
        elif next_char == "n":
            self.parse_literal("null")
            value = None
        else:
            raise self._unexpected_character()
        return value


class AbstractNode(object):

    def __init__(self, name, properties=None):
        self.name = name
        self.properties = properties or {}

    def __eq__(self, other):
        return self.name == other.name and self.properties == other.properties

    def __ne__(self, other):
        return self.name != other.name or self.properties != other.properties

    def __repr__(self):
        if self.properties:
            return "{0}({1}, {2})".format(self.__class__.__name__, repr(self.name), repr(self.properties))
        else:
            return "{0}({1})".format(self.__class__.__name__, repr(self.name))

    def __str__(self):
        if self.properties:
            return "({0} {1})".format(self.name, json.dumps(self.properties, separators=(",", ":")))
        else:
            return "({0})".format(self.name)


class AbstractRelationship(object):

    def __init__(self, start_node, type, properties, end_node):
        self.start_node = start_node
        self.type = type
        self.properties = properties or {}
        self.end_node = end_node

    def __eq__(self, other):
        return self.start_node == other.start_node and \
               self.type == other.type and \
               self.properties == other.properties and \
               self.end_node == other.end_node

    def __ne__(self, other):
        return self.start_node != other.start_node or \
               self.type != other.type or \
               self.properties != other.properties or \
               self.end_node != other.end_node

    def __str__(self):
        if self.properties:
            return "{0}-[:{1} {2}]->{3}".format(self.start_node, self.type, json.dumps(self.properties, separators=(",", ":")), self.end_node)
        else:
            return "{0}-[:{1}]->{2}".format(self.start_node, self.type, self.end_node)


class AbstractIndexEntry(object):

    def __init__(self, index_name, key, value, node):
        self.index_name = index_name
        self.key = key
        self.value = value
        self.node = node

    def __eq__(self, other):
        return self.index_name == other.index_name and \
               self.key == other.key and \
               self.value == other.value and \
               self.node == other.node

    def __ne__(self, other):
        return self.index_name != other.index_name or \
               self.key != other.key or \
               self.value != other.value or \
               self.node != other.node

    def __str__(self):
        return "|{0} {1}|=>{2}".format(self.index_name, json.dumps({self.key: self.value}, separators=(",", ":")), self.node)


def insert(graph_db, file):
    """ Insert Geoff data into a graph database.
    """
    Subgraph.load(file).insert_into(graph_db)


def merge(graph_db, file):
    """ Merge Geoff data into a graph database.
    """
    Subgraph.load(file).merge_into(graph_db)


def insert_xml(graph_db, file):
    """ Insert XML data into a graph database.
    """
    Subgraph.load_xml(file).insert_into(graph_db)


def merge_xml(graph_db, file):
    """ Merge XML data into a graph database.
    """
    Subgraph.load_xml(file).merge_into(graph_db)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" The ogm module provides Object to Graph Mapping features similar to ORM
facilities available for relational databases. All functionality is available
through the :py:class:`Store` class which is bound to a specific
:py:class:`neo4j.GraphDatabaseService` instance on creation.

Conceptually, a mapped object "owns" a single node within the graph along with
all of that node's outgoing relationships. These features are managed via a
pair of attributes called `__node__` and `__rel__` which store details of the
mapped node and the outgoing relationships respectively. The only specific
requirement for a mapped object is that it has a nullary constructor which can
be used to create new instances.

The `__node__` attribute holds a :py:class:`neo4j.Node` object which is the
node to which this object is mapped. If the attribute does not exist, or is
:py:const:`None`, the object is considered "unsaved".

The `__rel__` attribute holds a dictionary of outgoing relationship details.
Each key corresponds to a relationship type and each value to a list of
2-tuples representing the outgoing relationships of that type. Within each
2-tuple, the first value holds a dictionary of relationship properties (which
may be empty) and the second value holds the endpoint. The endpoint may be
either a :py:class:`neo4j.Node` instance or another mapped object. Any such
objects which are unsaved will be lazily saved as required by creation of the
relationship itself. The following data structure outline shows an example of
a `__rel__` attribute (where `alice` and `bob` represent other mapped objects::

    {
        "LIKES": [
            ({}, alice),
            ({"since": 1999}, bob)
        ]
    }

To manage relationships, use the :py:func:`Store.relate` and
:py:func:`Store.separate` methods. Neither method makes any calls to the
database and operates only on the local `__rel__` attribute. Changes must be
explicitly saved via one of the available save methods. The
:py:func:`Store.load_related` method loads all objects marked as related by
the `__rel__` attribute.

The code below shows an example of usage::

    from py2neo import neo4j, ogm

    class Person(object):

        def __init__(self, email=None, name=None, age=None):
            self.email = email
            self.name = name
            self.age = age

        def __str__(self):
            return self.name

    graph_db = neo4j.GraphDatabaseService()
    store = ogm.Store(graph_db)

    alice = Person("alice@example.com", "Alice", 34)
    store.save_unique("People", "email", alice.email, alice)

    bob = Person("bob@example.org", "Bob", 66)
    carol = Person("carol@example.net", "Carol", 42)
    store.relate(alice, "LIKES", bob)     # these relationships are not saved
    store.relate(alice, "LIKES", carol)   # until `alice` is saved
    store.save(alice)

    friends = store.load_related(alice, "LIKES", Person)
    print("Alice likes {0}".format(" and ".join(str(f) for f in friends)))

"""


from __future__ import unicode_literals

from . import neo4j


class NotSaved(ValueError):
    """ Raised when an object has not been saved but a bound node is required.
    """
    pass


class Store(object):

    def __init__(self, graph_db):
        self.graph_db = graph_db
        if self.graph_db.supports_optional_match:
            self.__delete_query = ("START a=node({A}) "
                                   "OPTIONAL MATCH a-[r]-b "
                                   "DELETE r, a")
        else:
            self.__delete_query = ("START a=node({A}) "
                                   "MATCH a-[r?]-b "
                                   "DELETE r, a")

    def _assert_saved(self, subj):
        try:
            node = subj.__node__
            if node is None:
                raise NotSaved(subj)
        except AttributeError:
            raise NotSaved(subj)

    def _get_node(self, endpoint):
        if isinstance(endpoint, neo4j.Node):
            return endpoint
        if not hasattr(endpoint, "__node__"):
            self.save(endpoint)
        return endpoint.__node__

    def _is_same(self, obj, endpoint):
        if isinstance(endpoint, neo4j.Node):
            if hasattr(obj, "__node__"):
                return endpoint == obj.__node__
            else:
                return False
        else:
            return endpoint is obj

    def is_saved(self, subj):
        """ Return :py:const:`True` if the object `subj` has been saved to
        the database, :py:const:`False` otherwise.

        :param subj: the object to test
        """
        return hasattr(subj, "__node__") and subj.__node__ is not None

    def relate(self, subj, rel_type, obj, properties=None):
        """ Define a relationship between `subj` and `obj` of type `rel_type`.
        This is a local operation only: nothing is saved to the database until
        a save method is called. Relationship properties may optionally be
        specified.

        :param subj: the object bound to the start of the relationship
        :param rel_type: the relationship type
        :param obj: the object bound to the end of the relationship
        :param properties: properties attached to the relationship (optional)
        """
        if not hasattr(subj, "__rel__"):
            subj.__rel__ = {}
        if rel_type not in subj.__rel__:
            subj.__rel__[rel_type] = []
        subj.__rel__[rel_type].append((properties or {}, obj))

    def separate(self, subj, rel_type, obj=None):
        """ Remove any relationship definitions which match the criteria
        specified. This is a local operation only: nothing is saved to the
        database until a save method is called. If no object is specified, all
        relationships of type `rel_type` are removed.

        :param subj: the object bound to the start of the relationship
        :param rel_type: the relationship type
        :param obj: the object bound to the end of the relationship (optional)
        """
        if not hasattr(subj, "__rel__"):
            return
        if rel_type not in subj.__rel__:
            return
        if obj is None:
            del subj.__rel__[rel_type]
        else:
            subj.__rel__[rel_type] = [
                (props, endpoint)
                for props, endpoint in subj.__rel__[rel_type]
                if not self._is_same(obj, endpoint)
            ]

    def load_related(self, subj, rel_type, cls):
        """ Load all nodes related to `subj` by a relationship of type
        `rel_type` into objects of type `cls`.

        :param subj: the object bound to the start of the relationship
        :param rel_type: the relationship type
        :param cls: the class to load all related objects into
        :return: list of `cls` instances
        """
        if not hasattr(subj, "__rel__"):
            return []
        if rel_type not in subj.__rel__:
            return []
        return [
            self.load(cls, self._get_node(endpoint))
            for rel_props, endpoint in subj.__rel__[rel_type]
        ]

    def load(self, cls, node):
        """ Load and return an object of type `cls` from database node `node`.

        :param cls: the class of the object to be returned
        :param node: the node from which to load object data
        :return: a `cls` instance
        """
        subj = cls()
        setattr(subj, "__node__", node)
        self.reload(subj)
        return subj

    def load_indexed(self, index_name, key, value, cls):
        """ Load zero or more indexed nodes from the database into a list of
        objects.

        :param index_name: the node index name
        :param key: the index key
        :param value: the index value
        :param cls: the class of the object to be returned
        :return: a list of `cls` instances
        """
        index = self.graph_db.get_index(neo4j.Node, index_name)
        nodes = index.get(key, value)
        return [self.load(cls, node) for node in nodes]

    def load_unique(self, index_name, key, value, cls):
        """ Load a uniquely indexed node from the database into an object.

        :param index_name: the node index name
        :param key: the index key
        :param value: the index value
        :param cls: the class of the object to be returned
        :return: as instance of `cls` containing the loaded data
        """
        index = self.graph_db.get_index(neo4j.Node, index_name)
        nodes = index.get(key, value)
        if not nodes:
            return None
        if len(nodes) > 1:
            raise LookupError("Multiple nodes match the given criteria; "
                              "consider using `load_all` instead.")
        return self.load(cls, nodes[0])

    def reload(self, subj):
        """ Reload properties and relationships from a database node into
        `subj`.

        :param subj: the object to reload
        :raise NotSaved: if `subj` is not linked to a database node
        """
        self._assert_saved(subj)
        # naively copy properties from node to object
        properties = subj.__node__.get_properties()
        for key in subj.__dict__:
            if not key.startswith("_") and key not in properties:
                setattr(subj, key, None)
        for key, value in properties.items():
            if not key.startswith("_"):
                setattr(subj, key, value)
        subj.__rel__ = {}
        for rel in subj.__node__.match():
            if rel.type not in subj.__rel__:
                subj.__rel__[rel.type] = []
            subj.__rel__[rel.type].append((rel.get_properties(), rel.end_node))

    def save(self, subj, node=None):
        """ Save an object to a database node.

        :param subj: the object to save
        :param node: the database node to save to (if omitted, will re-save to
            same node as previous save)
        """
        if node is not None:
            subj.__node__ = node
        # naively copy properties from object to node
        props = {}
        for key, value in subj.__dict__.items():
            if not key.startswith("_"):
                props[key] = value
        if hasattr(subj, "__node__"):
            subj.__node__.set_properties(props)
            query = neo4j.CypherQuery(self.graph_db, "START a=node({A}) "
                                                     "MATCH (a)-[r]->(b) "
                                                     "DELETE r")
            query.execute(A=subj.__node__._id)
        else:
            subj.__node__, = self.graph_db.create(props)
        # write rels
        if hasattr(subj, "__rel__"):
            batch = neo4j.WriteBatch(self.graph_db)
            for rel_type, rels in subj.__rel__.items():
                for rel_props, endpoint in rels:
                    end_node = self._get_node(endpoint)
                    if not neo4j.familiar(end_node, self.graph_db):
                        raise ValueError(end_node)
                    batch.create((subj.__node__, rel_type, end_node, rel_props))
            batch.run()
        return subj

    def save_indexed(self, index_name, key, value, *subj):
        """ Save one or more objects to the database, indexed under the
        supplied criteria.

        :param index_name: the node index name
        :param key: the index key
        :param value: the index value
        :param subj: one or more objects to save
        """
        index = self.graph_db.get_or_create_index(neo4j.Node, index_name)
        for subj in subj:
            index.add(key, value, self.save(self._get_node(subj)))

    def save_unique(self, index_name, key, value, subj):
        """ Save an object to the database, uniquely indexed under the
        supplied criteria.

        :param index_name: the node index name
        :param key: the index key
        :param value: the index value
        :param subj: the object to save
        """
        index = self.graph_db.get_or_create_index(neo4j.Node, index_name)
        node = index.get_or_create(key, value, {})
        self.save(subj, node)

    def delete(self, subj):
        """ Delete a saved object node from the database as well as all
        incoming and outgoing relationships.

        :param subj: the object to delete from the database
        :raise NotSaved: if `subj` is not linked to a database node
        """
        self._assert_saved(subj)
        node = subj.__node__
        del subj.__node__
        neo4j.CypherQuery(self.graph_db,
                          self.__delete_query).execute(A=node._id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Cypher Query Language module.
"""


from __future__ import unicode_literals

from collections import OrderedDict
import json

from .neo4j import DEFAULT_URI, CypherQuery, CypherError, ServiceRoot, Resource, _hydrated
from .util import deprecated, Record
from .packages.urimagic import URI


@deprecated("The cypher module is deprecated, use "
            "neo4j.CypherQuery instead")
class Metadata(object):
    """Metadata for query results.
    """

    #: List of column names
    columns = []

    def __init__(self, columns=None):
        self.columns = columns or []


@deprecated("The cypher module is deprecated, use "
            "neo4j.CypherQuery instead")
def execute(graph_db, query, params=None, row_handler=None,
            metadata_handler=None, error_handler=None):
    query = CypherQuery(graph_db, query)
    data, metadata = [], None
    try:
        results = query.execute(**params or {})
    except CypherError as err:
        if error_handler:
            error_handler(err.message, err.exception, err.stack_trace)
        else:
            raise
    else:
        metadata = Metadata(results.columns)
        if metadata_handler:
            metadata_handler(metadata)
        if row_handler:
            for record in results:
                row_handler(list(record))
            return data, metadata
        else:
            return [list(record) for record in results], metadata


def dumps(obj, separators=(", ", ": "), ensure_ascii=True):
    """ Dumps an object as a Cypher expression string.

    :param obj:
    :param separators:
    :return:
    """
    if isinstance(obj, dict):
        buffer = ["{"]
        link = ""
        for key, value in obj.items():
            buffer.append(link)
            if " " in key:
                buffer.append("`")
                buffer.append(key.replace("`", "``"))
                buffer.append("`")
            else:
                buffer.append(key)
            buffer.append(separators[1])
            buffer.append(dumps(value, separators=separators,
                                ensure_ascii=ensure_ascii))
            link = separators[0]
        buffer.append("}")
        return "".join(buffer)
    elif isinstance(obj, (tuple, set, list)):
        buffer = ["["]
        link = ""
        for value in obj:
            buffer.append(link)
            buffer.append(dumps(value, separators=separators,
                                ensure_ascii=ensure_ascii))
            link = separators[0]
        buffer.append("]")
        return "".join(buffer)
    else:
        return json.dumps(obj, ensure_ascii=ensure_ascii)
        
        
class Session(object):
    """ A Session is the base object from which Cypher transactions are
    created and is instantiated using a root service URI. If unspecified, this
    defaults to the `DEFAULT_URI`.

    ::

        >>> from py2neo import cypher
        >>> session = cypher.Session("http://arthur:excalibur@camelot:9999")

    """

    def __init__(self, uri=None):
        self._uri = URI(uri or DEFAULT_URI)
        if self._uri.user_info:
            service_root_uri = "{0}://{1}@{2}:{3}/".format(self._uri.scheme, self._uri.user_info, self._uri.host, self._uri.port)
        else:
            service_root_uri = "{0}://{1}:{2}/".format(self._uri.scheme, self._uri.host, self._uri.port)
        self._service_root = ServiceRoot.get_instance(service_root_uri)
        self._graph_db = self._service_root.graph_db
        try:
            self._transaction_uri = self._graph_db.__metadata__["transaction"]
        except KeyError:
            raise NotImplementedError("Cypher transactions are not supported "
                                      "by this server version")
        
    def create_transaction(self):
        """ Create a new transaction object.

        ::

            >>> from py2neo import cypher
            >>> session = cypher.Session()
            >>> tx = session.create_transaction()

        :return: new transaction object
        :rtype: Transaction
        """
        return Transaction(self._transaction_uri)

    def execute(self, statement, parameters=None):
        """ Execute a single statement and return the results.
        """
        tx = self.create_transaction()
        tx.append(statement, parameters)
        results = tx.execute()
        return results[0]

        
class Transaction(object):
    """ A transaction is a transient resource that allows multiple Cypher
    statements to be executed within a single server transaction.
    """

    def __init__(self, uri):
        self._begin = Resource(uri)
        self._begin_commit = Resource(uri + "/commit")
        self._execute = None
        self._commit = None
        self._clear()
        self._finished = False

    def _clear(self):
        self._statements = []

    def _assert_unfinished(self):
        if self._finished:
            raise TransactionFinished()

    @property
    def finished(self):
        """ Indicates whether or not this transaction has been completed or is
        still open.

        :return: :py:const:`True` if this transaction has finished,
                 :py:const:`False` otherwise
        """
        return self._finished

    def append(self, statement, parameters=None):
        """ Append a statement to the current queue of statements to be
        executed.

        :param statement: the statement to execute
        :param parameters: a dictionary of execution parameters
        """
        self._assert_unfinished()
        # OrderedDict is used here to avoid statement/parameters ordering bug
        self._statements.append(OrderedDict([
            ("statement", statement),
            ("parameters", dict(parameters or {})),
            ("resultDataContents", ["REST"]),
        ]))

    def _post(self, resource):
        self._assert_unfinished()
        rs = resource._post({"statements": self._statements})
        location = dict(rs.headers).get("location")
        if location:
            self._execute = Resource(location)
        j = rs.json
        rs.close()
        self._clear()
        if "commit" in j:
            self._commit = Resource(j["commit"])
        if "errors" in j:
            errors = j["errors"]
            if len(errors) >= 1:
                error = errors[0]
                raise TransactionError.new(error["code"], error["message"])
        return [
            [
                Record(result["columns"], _hydrated(r["rest"]))
                for r in result["data"]
            ]
            for result in j["results"]
        ]
        
    def execute(self):
        """ Send all pending statements to the server for execution, leaving
        the transaction open for further statements.

        :return: list of results from pending statements
        """
        return self._post(self._execute or self._begin)

    def commit(self):
        """ Send all pending statements to the server for execution and commit
        the transaction.

        :return: list of results from pending statements
        """
        try:
            return self._post(self._commit or self._begin_commit)
        finally:
            self._finished = True

    def rollback(self):
        """ Rollback the current transaction.
        """
        self._assert_unfinished()
        try:
            if self._execute:
                self._execute._delete()
        finally:
            self._finished = True


class TransactionError(Exception):
    """ Raised when an error occurs while processing a Cypher transaction.
    """

    @classmethod
    def new(cls, code, message):
        CustomError = type(str(code), (cls,), {})
        return CustomError(message)

    def __init__(self, message):
        Exception.__init__(self, message)


class TransactionFinished(Exception):
    """ Raised when actions are attempted against a finished Transaction.
    """

    def __init__(self):
        pass

    def __repr__(self):
        return "Transaction finished"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import unicode_literals

from .packages.jsonstream import assembled


__all__ = ["IndexTypeError", "ServerException", "ClientError", "ServerError",
           "CypherError", "BatchError"]


class IndexTypeError(TypeError):
    pass


class ServerException(object):

    def __init__(self, data):
        self._message = data.get("message")
        self._exception = data.get("exception")
        self._full_name = data.get("fullname")
        self._stack_trace = data.get("stacktrace")
        try:
            self._cause = ServerException(data["cause"])
        except KeyError:
            self._cause = None

    @property
    def message(self):
        return self._message

    @property
    def exception(self):
        return self._exception

    @property
    def full_name(self):
        return self._full_name

    @property
    def stack_trace(self):
        return self._stack_trace

    @property
    def cause(self):
        return self._cause


class ClientError(Exception):

    def __init__(self, response):
        assert response.status_code // 100 == 4
        try:
            self.__cause__ = response
        except TypeError:
            pass
        if response.is_json:
            self._server_exception = ServerException(assembled(response))
            Exception.__init__(self, self._server_exception.message)
        else:
            self._server_exception = None
            Exception.__init__(self, response.args[0])

    def __getattr__(self, item):
        try:
            return getattr(self._server_exception, item)
        except AttributeError:
            return getattr(self.__cause__, item)


class ServerError(Exception):

    def __init__(self, response):
        assert response.status_code // 100 == 5
        try:
            self.__cause__ = response
        except TypeError:
            pass
        # TODO: check for unhandled HTML errors (on 500)
        if response.is_json:
            self._server_exception = ServerException(assembled(response))
            Exception.__init__(self, self._server_exception.message)
        else:
            self._server_exception = None
            Exception.__init__(self, response.args[0])

    def __getattr__(self, item):
        try:
            return getattr(self._server_exception, item)
        except AttributeError:
            return getattr(self.__cause__, item)


class _FeatureError(Exception):

    def __init__(self, response):
        self._response = response
        Exception.__init__(self, self.message)

    @property
    def message(self):
        return self._response.message

    @property
    def exception(self):
        return self._response.exception

    @property
    def full_name(self):
        return self._response.full_name

    @property
    def stack_trace(self):
        return self._response.stack_trace

    @property
    def cause(self):
        return self._response.cause

    @property
    def request(self):
        return self._response.request

    @property
    def response(self):
        return self._response


class CypherError(_FeatureError):

    pass


class BatchError(_FeatureError):

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
An implementation of URI Templates from RFC 6570.
"""


from __future__ import unicode_literals

import re

from .rfc3986 import reserved, percent_encode, Part, URI


__all__ = ["URITemplate"]


class URITemplate(Part):
    """A URI Template is a compact sequence of characters for describing a
    range of Uniform Resource Identifiers through variable expansion.

    This class exposes a full implementation of RFC6570.
    """

    @classmethod
    def __cast(cls, obj):
        if obj is None:
            return cls(None)
        elif isinstance(obj, cls):
            return obj
        else:
            return cls(str(obj))

    class _Expander(object):

        _operators = set("+#./;?&")

        def __init__(self, values):
            self.values = values

        def collect(self, *keys):
            """ Fetch a list of all values matching the keys supplied,
            returning (key, value) pairs for each.
            """
            items = []
            for key in keys:
                if key.endswith("*"):
                    key, explode = key[:-1], True
                else:
                    explode = False
                if ":" in key:
                    key, max_length = key.partition(":")[0::2]
                    max_length = int(max_length)
                else:
                    max_length = None
                value = self.values.get(key)
                if isinstance(value, dict):
                    if not value:
                        items.append((key, None))
                    elif explode:
                        items.extend((key, _) for _ in value.items())
                    else:
                        items.append((key, value))
                elif isinstance(value, (tuple, list)):
                    if explode:
                        items.extend((key, _) for _ in value)
                    else:
                        items.append((key, list(value)))
                elif max_length is not None:
                    items.append((key, value[:max_length]))
                else:
                    items.append((key, value))
            return [(key, value) for key, value in items if value is not None]

        def _expand(self, expression, safe=None, prefix="", separator=",",
                    with_keys=False, trim_empty_equals=False):
            items = self.collect(*expression.split(","))
            encode = lambda x: percent_encode(x, safe=safe)
            for i, (key, value) in enumerate(items):
                if isinstance(value, tuple):
                    items[i] = "=".join(map(encode, value))
                else:
                    if isinstance(value, dict):
                        items[i] = ",".join(",".join(map(encode, item))
                                            for item in value.items())
                    elif isinstance(value, list):
                        items[i] = ",".join(map(encode, value))
                    else:
                        items[i] = encode(value)
                    if with_keys:
                        if items[i] is None or (items[i] == "" and
                                                trim_empty_equals):
                            items[i] = encode(key)
                        else:
                            items[i] = encode(key) + "=" + (items[i] or "")
            out = []
            for i, item in enumerate(items):
                out.append(prefix if i == 0 else separator)
                out.append(item)
            return "".join(out)

        def expand(self, expression):
            """ Dispatch to the correct expansion method.
            """
            if not expression:
                return ""
            if expression[0] in self._operators:
                operator, expression = expression[0], expression[1:]
                if operator == "+":
                    return self._expand(expression, reserved)
                elif operator == "#":
                    return self._expand(expression, reserved, prefix="#")
                elif operator == ".":
                    return self._expand(expression, prefix=".", separator=".")
                elif operator == "/":
                    return self._expand(expression, prefix="/", separator="/")
                elif operator == ";":
                    return self._expand(expression, prefix=";", separator=";",
                                        with_keys=True, trim_empty_equals=True)
                elif operator == "?":
                    return self._expand(expression, prefix="?", separator="&",
                                        with_keys=True)
                elif operator == "&":
                    return self._expand(expression, prefix="&", separator="&",
                                        with_keys=True)
            else:
                return self._expand(expression)

    _tokeniser = re.compile("(\{)([^{}]*)(\})")

    def __init__(self, template):
        super(URITemplate, self).__init__()
        self.__template = template

    def __eq__(self, other):
        other = self.__cast(other)
        return self.__template == other.__template

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.string)

    @property
    def string(self):
        if self.__template is None:
            return None
        return str(self.__template)

    def expand(self, **values):
        """ Expand into a URI using the values supplied
        """
        if self.__template is None:
            return URI(None)
        tokens = self._tokeniser.split(self.__template)
        expander = URITemplate._Expander(values)
        out = []
        while tokens:
            token = tokens.pop(0)
            if token == "{":
                expression = tokens.pop(0)
                tokens.pop(0)
                out.append(expander.expand(expression))
            else:
                out.append(token)
        return URI("".join(out))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import unicode_literals

from base64 import b64encode
import errno
try:
    from http.client import (BadStatusLine, CannotSendRequest,
                             HTTPConnection as _HTTPConnection,
                             HTTPSConnection as _HTTPSConnection,
                             HTTPException, ResponseNotReady,
                             responses)
except ImportError:
    from httplib import (BadStatusLine, CannotSendRequest,
                         HTTPConnection as _HTTPConnection,
                         HTTPSConnection as _HTTPSConnection,
                         HTTPException, ResponseNotReady,
                         responses)
import json
import logging
from os import strerror
from socket import error, gaierror, herror, timeout, IPPROTO_TCP, TCP_NODELAY
from threading import local
import sys
from xml.dom.minidom import parseString

from py2neo.packages.jsonstream import JSONStream
from py2neo.packages.urimagic import URI, URITemplate
from py2neo.packages.urimagic.kvlist import KeyValueList  # no point in another copy

from . import __version__
from .jsonencoder import JSONEncoder
from .numbers import *


__all__ = ["NetworkAddressError", "SocketError", "RedirectionError", "Request",
           "Response", "Redirection", "ClientError", "ServerError", "Resource",
           "ResourceTemplate", "get", "put", "post", "delete", "head"]


class HTTPConnection(_HTTPConnection):
    """ Patched class to avoid Nagle's algorithm:
    https://en.wikipedia.org/wiki/Nagle%27s_algorithm
    """

    def connect(self):
        _HTTPConnection.connect(self)
        self.sock.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)


class HTTPSConnection(_HTTPSConnection):
    """ Patched class to avoid Nagle's algorithm:
    https://en.wikipedia.org/wiki/Nagle%27s_algorithm
    """

    def connect(self):
        _HTTPSConnection.connect(self)
        self.sock.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)


connection_classes = {
    "http": HTTPConnection,
    "https": HTTPSConnection,
}

default_encoding = "ISO-8859-1"
default_chunk_size = 4096

log = logging.getLogger(__name__)
try:
    log.addHandler(logging.NullHandler())
except AttributeError:
    # Python 2.6
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
    log.addHandler(NullHandler())

redirects = {}

# Since the Python docs state that "symbols that are not used on the current
# platform are not defined by the module" we have to import error codes
# cautiously. Error numbers also vary across platforms so we cannot rely on
# raw numeric values either.
retry_codes = {}
if hasattr(errno, "EPIPE"):
    retry_codes[errno.EPIPE] = "broken pipe"
if hasattr(errno, "ENETRESET"):
    retry_codes[errno.ENETRESET] = "network reset"
if hasattr(errno, "ECONNABORTED"):
    retry_codes[errno.ECONNABORTED] = "connection aborted"
if hasattr(errno, "ECONNRESET"):
    retry_codes[errno.ECONNRESET] = "connection reset"

supports_buffering = ((2, 7) <= sys.version_info < (2, 8))


def user_agent(product=None):
    ua = []
    if product:
        if isinstance(product, (tuple, list)):
            ua.append("/".join(map(str, product)))
        else:
            ua.append(str(product))
    ua.append("HTTPStream/{0}".format(__version__))
    ua.append("Python/{0}.{1}.{2}-{3}-{4}".format(*sys.version_info))
    ua.append("({0})".format(sys.platform))
    return " ".join(ua)


class Loggable(object):

    def __init__(self, cls, message):
        log.error("!!! {0}: {1}".format(cls.__name__, message))


class NetworkAddressError(Loggable, IOError):

    def __init__(self, message, host_port=None):
        self._host_port = host_port
        IOError.__init__(self, message)
        Loggable.__init__(self, self.__class__, message)

    @property
    def host_port(self):
        return self._host_port


class SocketError(Loggable, IOError):

    def __init__(self, code, description, host_port=None):
        self._code = code
        self._description = description
        self._host_port = host_port
        IOError.__init__(self, description)
        Loggable.__init__(self, self.__class__, description)

    @property
    def code(self):
        return self._code

    @property
    def description(self):
        return self._description

    @property
    def host_port(self):
        return self._host_port


class RedirectionError(Loggable, HTTPException):

    def __init__(self, *args, **kwargs):
        HTTPException.__init__(self, *args, **kwargs)
        Loggable.__init__(self, self.__class__, args[0])


class ConnectionPuddle(local):
    """ A collection of HTTP/HTTPS connections to a single network location
    (i.e. host:port). Connections may be acquired and will be created if
    necessary; after use, these must be released.
    """

    def __init__(self, connection_class, host_port):
        local.__init__(self)
        self.__connection_class = connection_class
        self.__host_port = host_port
        self.__active = []
        self.__passive = []

    @property
    def host_port(self):
        return self.__host_port

    @property
    def connection_class(self):
        return self.__connection_class

    def __repr__(self):
        return "<{0}({1}) active={2} passive={3}>".format(
            self.__connection_class.__name__, repr(self.host_port),
            len(self.__active), len(self.__passive)
        )

    def __hash__(self):
        return hash((self.__connection_class, self.host_port))

    def __len__(self):
        return len(self.__active) + len(self.__passive)

    def acquire(self):
        if self.__passive:
            connection = self.__passive.pop()
        else:
            connection = self.__connection_class(self.host_port)
        self.__active.append(connection)
        return connection

    def release(self, connection):
        try:
            self.__active.remove(connection)
        except ValueError:
            pass
        if len(self.__passive) < 2:
            self.__passive.append(connection)
        else:
            connection.close()


class ConnectionPool(object):
    """ A collection of :py:class:`ConnectionPuddle` objects for various
    network locations.
    """

    _puddles = {}

    @classmethod
    def _get_puddle(cls, connection_class, host_port):
        key = (connection_class, host_port)
        if key not in cls._puddles:
            cls._puddles[key] = ConnectionPuddle(connection_class, host_port)
        return cls._puddles[key]

    @classmethod
    def acquire(cls, scheme, host_port):
        puddle = cls._get_puddle(connection_classes[scheme], host_port)
        return puddle.acquire()

    @classmethod
    def release(cls, connection):
        puddle = cls._get_puddle(connection.__class__, "{0}:{1}".format(
            connection.host, connection.port))
        puddle.release(connection)


def submit(method, uri, body, headers):
    """ Submit one HTTP request.
    """
    uri = URI(uri)
    headers["Host"] = uri.host_port
    if uri.user_info:
        credentials = uri.user_info.encode("UTF-8")
        value = "Basic " + b64encode(credentials).decode("ASCII")
        headers["Authorization"] = value
    try:
        http = ConnectionPool.acquire(uri.scheme, uri.host_port)
    except KeyError:
        raise ValueError("Unsupported URI scheme " + repr(uri.scheme))

    def send(reconnect=None):
        if reconnect:
            log.debug("<~> Reconnecting ({0})".format(reconnect))
            http.close()
            http.connect()
        if method in ("GET", "DELETE") and not body:
            log.info(">>> {0} {1}".format(method, uri))
        elif body:
            log.info(">>> {0} {1} [{2}]".format(method, uri, len(body)))
        else:
            log.info(">>> {0} {1} [0]".format(method, uri))
        if __debug__:
            for key, value in headers.items():
                log.debug(">>> {0}: {1}".format(key, value))
        http.request(method, uri.absolute_path_reference, body, headers)
        if supports_buffering:
            return http.getresponse(buffering=True)
        else:
            return http.getresponse()

    try:
        try:
            response = send()
        except BadStatusLine as err:
            if err.line == "''":
                response = send("peer closed connection")
            else:
                raise
        except ResponseNotReady:
            response = send("response not ready")
        except timeout:
            response = send("timeout")
        except error as err:
            if isinstance(err.args[0], tuple):
                code = err.args[0][0]
            else:
                code = err.args[0]
            if code in retry_codes:
                response = send(retry_codes[code])
            else:
                raise
    except (gaierror, herror) as err:
        raise NetworkAddressError(err.args[1], host_port=uri.host_port)
    except error as err:
        if isinstance(err.args[0], tuple):
            code, description = err.args[0]
        elif isinstance(err.args[0], int):
            code = err.args[0]
            try:
                description = strerror(code)
            except ValueError:
                description = None
        else:
            code, description = None, err.args[0]
        if code == 2:
            # Workaround for Linux bug with incorrect error message on
            # host resolution
            # ----
            # https://bugs.launchpad.net/ubuntu/+source/eglibc/+bug/1154599
            raise NetworkAddressError("Name or service not known",
                                      host_port=uri.host_port)
        else:
            raise SocketError(code, description, host_port=uri.host_port)
    else:
        return http, response


class Request(object):

    def __init__(self, method, uri, body=None, headers=None):
        if not uri:
            raise ValueError("No URI specified for request")
        #: HTTP method of this request
        self.method = method
        self.uri = uri
        self._headers = dict(headers or {})
        self.body = body

    @property
    def __uri__(self):
        return self.uri

    @property
    def uri(self):
        """ URI of the request.
        """
        return self._uri

    @uri.setter
    def uri(self, value):
        self._uri = URI(value)

    @property
    def body(self):
        """ Content of the request.
        """
        if isinstance(self._body, (dict, list, tuple)):
            return json.dumps(self._body, cls=JSONEncoder,
                              separators=(",", ":"))
        else:
            return self._body

    @body.setter
    def body(self, value):
        self._body = value

    @property
    def headers(self):
        """ Dictionary of headers attached to the request.
        """
        if isinstance(self._body, (dict, list, tuple)):
            self._headers.setdefault("Content-Type", "application/json")
        return self._headers

    def submit(self, redirect_limit=0, product=None, **response_kwargs):
        """ Submit this request and return a
        :py:class:`Response <httpstream.Response>` object.
        """
        uri = URI(self.uri)
        headers = dict(self.headers)
        headers.setdefault("User-Agent", user_agent(product))
        while True:
            http, rs = submit(self.method, uri, self.body, headers)
            status_class = rs.status // 100
            if status_class == 3:
                redirection = Redirection(http, uri, self, rs,
                                          **response_kwargs)
                if redirect_limit:
                    redirect_limit -= 1
                    location = URI.resolve(uri, rs.getheader("Location"))
                    if location == uri:
                        raise RedirectionError("Circular redirection")
                    if rs.status in (MOVED_PERMANENTLY, PERMANENT_REDIRECT):
                        redirects[uri] = location
                    uri = location
                    redirection.close()
                else:
                    return redirection
            elif status_class == 4:
                raise ClientError(http, uri, self, rs, **response_kwargs)
            elif status_class == 5:
                raise ServerError(http, uri, self, rs, **response_kwargs)
            else:
                return Response(http, uri, self, rs, **response_kwargs)


class Response(object):
    """ File-like object allowing consumption of an HTTP response.
    """

    def __init__(self, http, uri, request, response, **kwargs):
        self._http = http
        self._uri = URI(uri)
        self._request = request
        self._response = response
        self._reason = kwargs.get("reason")
        self.__headers = KeyValueList(self._response.getheaders())
        #: Default chunk size for this response
        self.chunk_size = kwargs.get("chunk_size", default_chunk_size)
        log.info("<<< {0}".format(self))
        if __debug__:
            for key, value in self._response.getheaders():
                log.debug("<<< {0}: {1}".format(key, value))

    def __del__(self):
        self.close()

    def __repr__(self):
        if self.is_chunked:
            return "{0} {1} [chunked]".format(self.status_code, self.reason)
        else:
            return "{0} {1} [{2}]".format(self.status_code, self.reason,
                                          self.content_length)

    def __getitem__(self, key):
        if not self._response:
            return None
        return self._response.getheader(key)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    @property
    def closed(self):
        """ Indicates whether or not the response is closed.
        """
        return not bool(self._http)

    def close(self):
        """ Close the response, discarding all remaining content and releasing
        the underlying connection object.
        """
        if self._http:
            try:
                self._response.read()
            except HTTPException:
                pass
            ConnectionPool.release(self._http)
            self._http = None

    @property
    def __uri__(self):
        return self._uri

    @property
    def uri(self):
        """ The URI from which the response came.
        """
        return self._uri

    @property
    def request(self):
        """ The :py:class:`Request` object which preceded this response.
        """
        return self._request

    @property
    def status_code(self):
        """ The status code of the response
        """
        return self._response.status

    @property
    def reason(self):
        """ The reason phrase attached to this response.
        """
        if self._reason:
            return self._reason
        else:
            return responses[self.status_code]

    @property
    def headers(self):
        """ The response headers.
        """
        return self.__headers

    @property
    def content_length(self):
        """ The length of content as provided by the `Content-Length` header
        field. If the content is chunked, this returns :py:const:`None`.
        """
        if not self.is_chunked:
            return int(self._response.getheader("Content-Length", 0))

    @property
    def content_type(self):
        """ The type of content as provided by the `Content-Type` header field.
        """
        try:
            content_type = [
                _.strip()
                for _ in self._response.getheader("Content-Type").split(";")
            ]
        except AttributeError:
            return None
        return content_type[0]

    @property
    def encoding(self):
        """ The content character set encoding.
        """
        try:
            content_type = dict(
                _.strip().partition("=")[0::2]
                for _ in self._response.getheader("Content-Type").split(";")
            )
        except AttributeError:
            return default_encoding
        return content_type.get("charset", default_encoding)

    @property
    def is_chunked(self):
        """ Indicates whether or not the content is chunked.
        """
        return self._response.getheader("Transfer-Encoding") == "chunked"

    @property
    def is_json(self):
        """ Indicates whether or not the content is JSON.
        """
        return self.content_type in ("application/json",
                                     "application/x-javascript",
                                     "text/javascript",
                                     "text/json")

    @property
    def json(self):
        """ Fetch all content, decoding from JSON and returning the decoded
        value.
        """
        if not self.is_json:
            raise TypeError("Content is not JSON")
        return json.loads(self.read().decode(self.encoding))

    @property
    def is_text(self):
        """ Indicates whether or not the content is plain text.
        """
        return self.content_type.partition("/")[0] == "text"

    @property
    def text(self):
        """ Fetches all content as a string.
        """
        if not self.is_text and not self.is_json and not self.is_xml:
            raise TypeError("Content is not text")
        return self.read().decode(self.encoding)

    @property
    def is_tsj(self):
        """ Indicates whether or not the content is tab-separated JSON.
        """
        return self.content_type == "text/x-tab-separated-json"

    @property
    def tsj(self):
        """ Fetches all content, decoding from tab-separated JSON and returning
        the decoded values.
        """
        if not self.is_tsj:
            raise TypeError("Content is not tab-separated JSON")
        return [
            [json.loads(value) for value in line.split("\t")]
            for line in self.read().decode(self.encoding).splitlines()
        ]

    @property
    def is_xml(self):
        """ Indicates whether or not the content is XML.
        """
        return self.content_type == "application/xml"

    @property
    def dom(self):
        """ Fetches all content, decoding from XML and returning as a DOM
        object.
        """
        if not self.is_xml:
            raise TypeError("Content is not XML")
        return parseString(self.read().decode(self.encoding))

    @property
    def content(self):
        """ Fetch all content, returning a value appropriate for the content
        type.
        """
        if self.status_code == NO_CONTENT:
            return None
        elif self.is_json:
            return self.json
        elif self.is_tsj:
            return self.tsj
        elif self.is_text:
            return self.text
        else:
            return self.read()

    def read(self, size=None):
        """ Fetch some or all of the response content, returning as a bytearray.
        """
        completed = False
        try:
            if size is None:
                data = self._response.read()
                completed = True
            else:
                data = self._response.read(size)
                completed = bool(size and not data)
            return data
        finally:
            if completed:
                self.close()

    def iter_chunks(self, chunk_size=None):
        """ Iterate through the content as chunks of text. Chunk sizes may vary
        slightly from that specified due to multi-byte characters. If no chunk
        size is specified, a default of 4096 is used.
        """
        try:
            if not chunk_size:
                chunk_size = self.chunk_size
            pending = bytearray()
            data = True
            while data:
                data = self.read(chunk_size)
                pending.extend(data)
                decoded = None
                while data and not decoded:
                    try:
                        decoded = pending.decode(self.encoding)
                    except UnicodeDecodeError:
                        data = self.read(1)
                        pending.extend(data)
                    else:
                        del pending[:]
                        yield decoded
        finally:
            self.close()

    def iter_json(self):
        """ Iterate through the content as individual JSON values.
        """
        return iter(JSONStream(self.iter_chunks()))

    def iter_lines(self, keep_ends=False):
        """ Iterate through the content as lines of text.
        """
        data = ""
        for chunk in self.iter_chunks():
            data += chunk
            while "\r" in data or "\n" in data:
                cr, lf = data.find("\r"), data.find("\n")
                if cr >= 0 and lf == cr + 1:
                    eol_pos, eol_len = cr, 2
                else:
                    if cr >= 0 and lf >= 0:
                        eol_pos = min(cr, lf)
                    else:
                        eol_pos = cr if cr >= 0 else lf
                    eol_len = 1
                x = eol_pos + eol_len
                if keep_ends:
                    line, data = data[:x], data[x:]
                else:
                    line, data = data[:eol_pos], data[x:]
                yield line
        if data:
            yield data

    def iter_tsj(self):
        """ Iterate through the content as lines of tab-separated JSON.
        """
        for line in self.iter_lines():
            yield [json.loads(value) for value in line.split("\t")]

    def __iter__(self):
        if self.status_code == NO_CONTENT:
            return iter([])
        elif self.is_json:
            return self.iter_json()
        elif self.is_tsj:
            return self.iter_tsj()
        elif self.is_text:
            return self.iter_lines()
        else:
            return self.iter_chunks()


class Redirection(Response):

    def __init__(self, http, uri, request, response, **kwargs):
        assert response.status // 100 == 3
        Response.__init__(self, http, uri, request, response, **kwargs)


class ClientError(Exception, Response):

    def __init__(self, http, uri, request, response, **kwargs):
        assert response.status // 100 == 4
        Response.__init__(self, http, uri, request, response, **kwargs)
        Exception.__init__(self, self.reason)


class ServerError(Exception, Response):

    def __init__(self, http, uri, request, response, **kwargs):
        assert response.status // 100 == 5
        Response.__init__(self, http, uri, request, response, **kwargs)
        Exception.__init__(self, self.reason)


class Resource(object):
    """ A web resource identified by a URI.
    """

    def __init__(self, uri):
        self._uri = URI(uri)

    def __str__(self):
        return "<{0}>".format(str(self._uri))

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__,
                                 repr(self._uri.string))

    def __eq__(self, other):
        """ Determine equality of two objects based on URI.
        """
        return self._uri == other._uri

    def __ne__(self, other):
        """ Determine inequality of two objects based on URI.
        """
        return self._uri != other._uri

    def __bool__(self):
        return bool(self._uri)

    def __nonzero__(self):
        return bool(self._uri)

    @property
    def __uri__(self):
        return self._uri

    @property
    def uri(self):
        """ The URI of this resource.
        """
        return self._uri

    def resolve(self, reference, strict=True):
        """ Resolve a URI reference against the URI for this resource,
        returning a new resource represented by the new target URI.
        """
        return Resource(self._uri.resolve(reference, strict))

    def get(self, headers=None, redirect_limit=5, **kwargs):
        """ Issue a ``GET`` request to this resource.

        :param headers: headers to be included in the request (optional)
        :type headers: dict
        :param redirect_limit: maximum number of redirects to be handled
            automatically (optional, default=5)
        :param product: name or (name, version) tuple for the client product to
            be listed in the ``User-Agent`` header (optional)
        :param chunk_size: number of bytes to retrieve per chunk (optional,
            default=4096)
        :return: file-like :py:class:`Response <httpstream.http.Response>`
            object from which content can be read
        """
        rq = Request("GET", self._uri, None, headers)
        return rq.submit(redirect_limit=redirect_limit, **kwargs)

    def put(self, body=None, headers=None, **kwargs):
        """ Issue a ``PUT`` request to this resource.
        """
        rq = Request("PUT", self._uri, body, headers)
        return rq.submit(**kwargs)

    def post(self, body=None, headers=None, **kwargs):
        """ Issue a ``POST`` request to this resource.
        """
        rq = Request("POST", self._uri, body, headers)
        return rq.submit(**kwargs)

    def delete(self, headers=None, **kwargs):
        """ Issue a ``DELETE`` request to this resource.
        """
        rq = Request("DELETE", self._uri, None, headers)
        return rq.submit(**kwargs)

    def head(self, headers=None, redirect_limit=5, **kwargs):
        """ Issue a ``HEAD`` request to this resource.
        """
        rq = Request("HEAD", self._uri, None, headers)
        return rq.submit(redirect_limit=redirect_limit, **kwargs)


class ResourceTemplate(object):

    def __init__(self, uri_template):
        if isinstance(uri_template, URITemplate):
            self._uri_template = uri_template
        else:
            self._uri_template = URITemplate(uri_template)

    def __str__(self):
        return "<{0}>".format(str(self._uri_template))

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__,
                                 repr(self._uri_template.string))

    def __eq__(self, other):
        return self._uri_template == other._uri_template

    def __ne__(self, other):
        return self._uri_template != other._uri_template

    def __bool__(self):
        return bool(self._uri_template)

    def __nonzero__(self):
        return bool(self._uri_template)

    @property
    def uri_template(self):
        """ The URI template string of this resource template.
        """
        return self._uri_template

    def expand(self, **values):
        """ Expand this template into a full URI using the values provided.
        """
        return Resource(self._uri_template.expand(**values))


def get(uri, headers=None, redirect_limit=5, **kwargs):
    return Resource(uri).get(headers, redirect_limit, **kwargs)


def put(uri, body=None, headers=None, **kwargs):
    return Resource(uri).put(body, headers, **kwargs)


def post(uri, body=None, headers=None, **kwargs):
    return Resource(uri).post(body, headers, **kwargs)


def delete(uri, headers=None, **kwargs):
    return Resource(uri).delete(headers, **kwargs)


def head(uri, headers=None, redirect_limit=5, **kwargs):
    return Resource(uri).head(headers, redirect_limit, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility module
"""


from __future__ import unicode_literals

from itertools import cycle, islice
import re
import warnings


__all__ = ["numberise", "compact", "flatten", "round_robin", "deprecated",
           "version_tuple", "is_collection", "has_all", "ustr", "pendulate",
           "Record"]


def numberise(n):
    """ Convert a value to an integer if possible. If not, simply return
        the input value.
    """
    if n == "NaN":
        return None
    try:
        return int(n)
    except ValueError:
        return n


def compact(obj):
    """ Return a copy of an object with all :py:const:`None` values removed.
    """
    if isinstance(obj, dict):
        return dict((key, value) for key, value in obj.items() if value is not None)
    else:
        return obj.__class__(value for value in obj if value is not None)


def flatten(*values):
    for value in values:
        if hasattr(value, "__iter__"):
            for val in value:
                yield val
        else:
            yield value


def round_robin(*iterables):
    """ Cycle through a number of iterables, returning
        the next item from each in turn.

        round_robin('ABC', 'D', 'EF') --> A D E B F C

        Original recipe credited to George Sakkis
        Python 2/3 cross-compatibility tweak by Nigel Small
    """
    pending = len(iterables)
    nexts = cycle(iter(it) for it in iterables)
    while pending:
        try:
            for n in nexts:
                yield next(n)
        except StopIteration:
            pending -= 1
            nexts = cycle(islice(nexts, pending))


def deprecated(message):
    """ Decorator for deprecating functions and methods.

    ::

        @deprecated("'foo' has been deprecated in favour of 'bar'")
        def foo(x):
            pass

    """
    def f__(f):
        def f_(*args, **kwargs):
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            return f(*args, **kwargs)
        f_.__name__ = f.__name__
        f_.__doc__ = f.__doc__
        f_.__dict__.update(f.__dict__)
        return f_
    return f__


VERSION = re.compile("(\d+\.\d+(\.\d+)?)")


def version_tuple(string):
    numbers = VERSION.match(string)
    if numbers:
        version = [int(n) for n in numbers.group(0).split(".")]
        extra = string[len(numbers.group(0)):]
        while extra.startswith(".") or extra.startswith("-"):
            extra = extra[1:]
    else:
        version = []
        extra = string
    while len(version) < 3:
        version += [0]
    version += [extra]
    return tuple(version)


def is_collection(obj):
    """ Returns true for any iterable which is not a string or byte sequence.
    """
    try:
        if isinstance(obj, unicode):
            return False
    except NameError:
        pass
    if isinstance(obj, bytes):
        return False
    try:
        iter(obj)
    except TypeError:
        return False
    try:
        hasattr(None, obj)
    except TypeError:
        return True
    return False


has_all = lambda iterable, items: all(item in iterable for item in items)


try:
    unicode
except NameError:
    # Python 3
    def ustr(s, encoding="utf-8"):
        if isinstance(s, str):
            return s
        try:
            return s.decode(encoding)
        except AttributeError:
            return str(s)
else:
    # Python 2
    def ustr(s, encoding="utf-8"):
        if isinstance(s, str):
            return s.decode(encoding)
        else:
            return unicode(s)


def pendulate(collection):
    count = len(collection)
    for i in range(count):
        if i % 2 == 0:
            index = i // 2
        else:
            index = count - ((i + 1) // 2)
        yield index, collection[index]


class Record(object):
    """ A single row of a Cypher execution result, holding a sequence of named
    values.
    """

    def __init__(self, columns, values):
        self._columns = tuple(columns)
        self._column_indexes = dict((b, a) for a, b in enumerate(columns))
        self._values = tuple(values)
    
    def __repr__(self):
        return "Record(columns={0}, values={1})".format(self._columns, self._values)
        
    def __getattr__(self, attr):
        return self._values[self._column_indexes[attr]]
        
    def __getitem__(self, item):
        if isinstance(item, (int, slice)):
            return self._values[item]
        else:
            return self._values[self._column_indexes[item]]

    def __len__(self):
        return len(self._columns)

    @property
    def columns(self):
        """ The column names defined for this record.

        :return: tuple of column names
        """
        return self._columns

    @property
    def values(self):
        """ The values stored in this record.

        :return: tuple of values
        """
        return self._values



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Command line swiss army knife for Neo4j.
"""


from __future__ import unicode_literals

import codecs
import json
import locale
import logging
import os
import readline
import sys

from . import __version__, __copyright__, neo4j, geoff
from .exceptions import CypherError
from .util import ustr
from .xmlutil import xml_to_cypher, xml_to_geoff


PY3 = sys.version > '3'
SCRIPT_NAME = "neotool"

NEOTOOL_HELP = """\
Usage:
  {script} <options> <command> <args>
Options:
  -h/--help [<command>]           Show tool usage
  -v/--version                    Show tool version
  -c/--copyright                  Show tool copyright
  -S/--scheme <scheme>            Set database scheme
  -H/--host <host>                Set database host
  -P/--port <port>                Set database port
  -U/--user <user>                Set HTTP basic auth user
  -W/--password <password>        Set HTTP basic auth password
Commands:
  clear                           Clear all nodes and relationships
  cypher <query>                  Execute Cypher and output as text
  cypher-csv <query>              Execute Cypher and output as CSV
  cypher-geoff <query>            Execute Cypher and output as Geoff
  cypher-json <query>             Execute Cypher and output as JSON
  cypher-tsv <query>              Execute Cypher and output as TSV
  geoff-insert <file>             Insert Geoff data
  geoff-merge <file>              Merge Geoff data
  shell                           Start an interactive shell
  xml-cypher <file> [<xmlns>...]  Convert XML data to Cypher CREATE statement
  xml-geoff <file> [<xmlns>...]   Convert XML data to Geoff data
"""

SHELL_HELP = """\
The Neotool Shell allows you to run Cypher queries from an interactive prompt.
Queries may be entered directly or run from files using the EXECUTE command. To
quit the shell, press Ctrl+D or use the EXIT command.

Commands available:

    ADD PARAM[ETER]S <json-file>
    CLEAR
    EXECUTE <cypher-file>
    EXIT
    HELP
    LOOKUP
    REMOVE PARAM[ETER]S
    SHOW PARAM[ETER]S
    VERSION

If a numeric value is entered, details of the node with that ID are displayed.
Any other statements are executed as Cypher.
"""

if not PY3:
    _stdin = sys.stdin
    preferred_encoding = locale.getpreferredencoding()
    sys.stdin = codecs.getreader(preferred_encoding)(sys.stdin)
    sys.stdout = codecs.getwriter(preferred_encoding)(sys.stdout)
    sys.stderr = codecs.getwriter(preferred_encoding)(sys.stderr)


class ResultWriter(object):

    def __init__(self, out=None):
        self.out = out or sys.stdout

    @classmethod
    def _stringify(cls, value, quoted=False):
        if value is None:
            if quoted:
                return "null"
            else:
                return ""
        elif isinstance(value, list):
            out = " ".join(
                cls._stringify(item, quoted=False)
                for item in value
            )
            if quoted:
                out = json.dumps(out, separators=(',', ':'), ensure_ascii=False)
        else:
            if quoted:
                try:
                    out = json.dumps(value, ensure_ascii=False)
                except TypeError:
                    out = json.dumps(ustr(value), ensure_ascii=False)
            else:
                out = ustr(value)
        return out

    @classmethod
    def _jsonify(cls, value):
        if isinstance(value, list):
            out = "[" + ", ".join(cls._jsonify(i) for i in value) + "]"
        elif hasattr(value, "__uri__") and hasattr(value, "_properties"):
            metadata = {
                "uri": ustr(value.__uri__),
                "properties": value._properties,
            }
            try:
                metadata.update({
                    "start": ustr(value.start_node.__uri__),
                    "type": ustr(value.type),
                    "end": ustr(value.end_node.__uri__),
                })
            except AttributeError:
                pass
            out = json.dumps(metadata, ensure_ascii=False)
        else:
            out = json.dumps(value, ensure_ascii=False)
        return out

    def write_delimited(self, record_set, **kwargs):
        field_delimiter = kwargs.get("field_delimiter", "\t")
        self.out.write(field_delimiter.join([
            json.dumps(column, ensure_ascii=False)
            for column in record_set.columns
        ]))
        self.out.write("\n")
        for row in record_set:
            self.out.write(field_delimiter.join([
                ResultWriter._stringify(value, quoted=True)
                for value in row
            ]))
            self.out.write("\n")

    def write_geoff(self, record_set):
        nodes = set()
        rels = set()

        def update_descriptors(value):
            if isinstance(value, list):
                for item in value:
                    update_descriptors(item)
            elif hasattr(value, "__uri__") and hasattr(value, "_properties"):
                if hasattr(value, "type"):
                    rels.add(value)
                else:
                    nodes.add(value)

        for row in record_set:
            for i in range(len(row)):
                update_descriptors(row[i])
        for node in sorted(nodes, key=lambda x: x._id):
            self.out.write(ustr(node))
            self.out.write("\n")
        for rel in sorted(rels, key=lambda x: x._id):
            self.out.write(ustr(rel))
            self.out.write("\n")

    def write_json(self, record_set):
        columns = [json.dumps(column, ensure_ascii=False) for column in record_set.columns]
        row_count = 0
        self.out.write("[")
        for row in record_set:
            row_count += 1
            if row_count > 1:
                self.out.write(", ")
            self.out.write("{" + ", ".join([
                columns[i] + ": " + ResultWriter._jsonify(row[i])
                for i in range(len(row))
            ]) + "}")
        self.out.write("]")

    def write_text(self, record_set):
        columns = record_set.columns
        column_widths = [len(column) for column in columns]
        data = [
            [
                ResultWriter._stringify(value)
                for value in row
            ]
            for row in record_set
        ]
        for row in data:
            column_widths = [
                max(column_widths[i], len(value))
                for i, value in enumerate(row)
            ]
        self.out.write(" " + " | ".join(
            columns[i].ljust(column_widths[i])
            for i, column in enumerate(columns)
        ) + " \n")
        self.out.write("-" + "-+-".join(
            "".ljust(column_widths[i], "-")
            for i, column in enumerate(columns)
        ) + "-\n")
        for row in data:
            self.out.write(" " + " | ".join([
                value.ljust(column_widths[i])
                for i, value in enumerate(row)
            ]) + " \n")
        if len(data) == 1:
            self.out.write("(1 row)\n\n")
        else:
            self.out.write("({0} rows)\n\n".format(len(data)))

    formats = {
        "csv": (write_delimited, {"field_delimiter": ","}),
        "geoff": (write_geoff, {}),
        "json": (write_json, {}),
        "text": (write_text, {}),
        "tsv": (write_delimited, {"field_delimiter": "\t"}),
    }

    def write(self, format_, record_set):
        try:
            method, kwargs = self.formats[format_]
        except KeyError:
            raise ValueError("Unknown format {0}".format(repr(format_)))
        if kwargs:
            method(self, record_set, **kwargs)
        else:
            method(self, record_set)


class Tool(object):

    def __init__(self, in_=None, out=None, err=None):
        self._scheme = "http"
        self._host = "localhost"
        self._port = 7474
        self._user = None
        self._password = None
        self._in = in_ or sys.stdin
        self._out = out or sys.stdout
        self._err = err or sys.stderr
        self._script = None

    @property
    def _graph_db(self):
        host_port = "{0}:{1}".format(self._host, self._port)
        uri = "{0}://{1}".format(self._scheme, host_port)
        if self._user and self._password:
            neo4j.authenticate(host_port, self._user, self._password)
        return neo4j.ServiceRoot(uri).graph_db

    def _version(self):
        """ Show tool version
        """
        self._out.write("{0} (py2neo/{1})\n".format(SCRIPT_NAME, __version__))

    def _copyright(self):
        """ Show tool copyright
        """
        self._out.write("(C) Copyright {0}\n".format(__copyright__))

    def _help(self):
        """ Show tool usage
        """
        script = self._script.split(os.sep)[-1].rstrip(".py")
        self._version()
        self._out.write(NEOTOOL_HELP.format(script=script))
        self._copyright()

    def do(self, argv):
        self._script = argv.pop(0)
        command = None
        while not command:
            try:
                arg = argv.pop(0)
            except IndexError:
                self._help()
                sys.exit(0)
            if arg.startswith("-"):
                if arg in ("-h", "--help"):
                    self._help()
                    sys.exit(0)
                elif arg in ("-v", "--version"):
                    self._version()
                    sys.exit(0)
                elif arg in ("-c", "--copyright"):
                    self._copyright()
                    sys.exit(0)
                elif arg in ("-S", "--scheme"):
                    self._scheme = argv.pop(0)
                elif arg in ("-H", "--host"):
                    self._host = argv.pop(0)
                elif arg in ("-P", "--port"):
                    self._port = int(argv.pop(0))
                elif arg in ("-U", "--user"):
                    self._user = argv.pop(0)
                elif arg in ("-W", "--password"):
                    self._password = argv.pop(0)
                elif arg in ("-d", "--debug"):
                    logging.basicConfig(level=logging.DEBUG)
                elif arg in ("-i", "--info"):
                    logging.basicConfig(level=logging.INFO)
                else:
                    raise ValueError("Unknown option {0}".format(repr(arg)))
            else:
                command = arg
        try:
            method = getattr(self, command.replace("-", "_"))
        except AttributeError:
            raise ValueError("Unknown command {0}".format(repr(command)))
        args, kwargs = [], {}
        for arg in argv:
            if " " not in arg and "=" in arg:
                key, value = arg.partition("=")[0::2]
                kwargs[key] = value
            else:
                args.append(arg)
        try:
            method(*args, **kwargs)
        except TypeError as e:
            raise ValueError("Incorrect usage: {0}".format(e))

    def clear(self):
        """ Clear all nodes and relationships.
        """
        self._graph_db.clear()

    def _cypher(self, format_, query, params=None):
        if query == "-":
            query = self._in.read()
        record_set = neo4j.CypherQuery(self._graph_db, query).execute(**params or {})
        writer = ResultWriter(self._out)
        writer.write(format_, record_set)

    def cypher(self, query=None, **params):
        """ Execute Cypher query and output as text.
        """
        if not query:
            query = sys.stdin.read()
            sys.stdin.close()
        self._cypher("text", query, params)

    def cypher_csv(self, query=None, **params):
        """ Execute Cypher query and output as comma separated values
        """
        if not query:
            query = sys.stdin.read()
            sys.stdin.close()
        self._cypher("csv", query, params)

    def cypher_geoff(self, query=None, **params):
        """ Execute Cypher query and output as Geoff
        """
        if not query:
            query = sys.stdin.read()
            sys.stdin.close()
        self._cypher("geoff", query, params)

    def cypher_json(self, query=None, **params):
        """ Execute Cypher query and output as JSON
        """
        if not query:
            query = sys.stdin.read()
            sys.stdin.close()
        self._cypher("json", query, params)

    def cypher_tsv(self, query=None, **params):
        """ Execute Cypher query and output as tab separated values
        """
        if not query:
            query = sys.stdin.read()
            sys.stdin.close()
        self._cypher("tsv", query, params)

    def dump_geoff(self, **params):
        """ Dump all entities from the database and output as Geoff
        """
        query = "START n=node(*) RETURN n"
        self._cypher("geoff", query, params)
        query = "START r=rel(*) RETURN r"
        self._cypher("geoff", query, params)

    def _geoff_write(self, params):
        for key, value in params.items():
            self._out.write(key)
            self._out.write("\t")
            self._out.write(ustr(value))
            self._out.write("\n")

    def geoff_insert(self, file_name=None):
        """ Insert Geoff data
        """
        if file_name:
            file = codecs.open(file_name, encoding="utf-8")
        else:
            file = sys.stdin
        params = geoff.Subgraph.load(file).insert_into(self._graph_db)
        self._geoff_write(params)

    def geoff_merge(self, file_name=None):
        """ Merge Geoff data
        """
        if file_name:
            file = codecs.open(file_name, encoding="utf-8")
        else:
            file = sys.stdin
        params = geoff.Subgraph.load(file).merge_into(self._graph_db)
        self._geoff_write(params)

    def xml_cypher(self, file_name=None, **prefixes):
        """ Convert XML data to Cypher CREATE statement.
        """
        if file_name:
            file = codecs.open(file_name, encoding="utf-8")
        else:
            file = sys.stdin
        src = file.read().encode("utf-8")
        file.close()
        self._out.write(xml_to_cypher(src, prefixes=prefixes))
        self._out.write("\n")

    def xml_geoff(self, file_name=None, **prefixes):
        """ Convert XML data to Geoff.
        """
        if file_name:
            file = codecs.open(file_name, encoding="utf-8")
        else:
            file = sys.stdin
        src = file.read().encode("utf-8")
        file.close()
        self._out.write(xml_to_geoff(src, prefixes=prefixes))
        self._out.write("\n")
        
    def shell(self):
        Shell(self._graph_db).repl()


class CommandLine(object):

    def __init__(self, text):
        self.text = text.strip()
    
    def __bool__(self):
        return bool(self.text)
    
    def __nonzero__(self):
        return bool(self.text)
    
    def peek(self):
        bits = self.text.split(None, 1)
        if not bits:
            return None
        return bits[0]
    
    def pop(self):
        bits = self.text.split(None, 1)
        if not bits:
            return None
        try:
            self.text = bits[1]
        except IndexError:
            self.text = ""
        return bits[0]


if PY3:
    def get_input(prompt):
        return input(prompt)
else:
    def get_input(prompt):
        sys.stdin = _stdin
        return raw_input(prompt).decode(sys.stdin.encoding)


class Shell(object):

    def __init__(self, graph_db):
        self.graph_db = graph_db
        self.lang = "cypher"
        self.format = "text"
        self.param_sets = []

    @property
    def prompt(self):
        if self.param_sets:
            return "\x1b[32;1m{0}/{1}\x1b[36;1m[{2}]\x1b[32;1m>\x1b[0m ".format(self.graph_db.service_root.__uri__.host_port, self.lang, len(self.param_sets))
        else:
            return "\x1b[32;1m{0}/{1}>\x1b[0m ".format(self.graph_db.service_root.__uri__.host_port, self.lang)

    def repl(self):
        print("Neotool Shell (py2neo/{0} Python/{1}.{2}.{3}-{4}-{5})".format(__version__, *sys.version_info))
        print("Copyright 2013-2014, Nigel Small")
        print("")
        try:
            while True:
                line = get_input(self.prompt)
                self.execute(line)
        except EOFError:
            print("⌁")
        except StopIteration:
            pass

    def _pop_command_and_argument(self, line):
        return line.pop(), line.pop()

    def execute(self, line):
        line = CommandLine(line)
        if not line:
            return
        command = line.peek().upper()
        if all(ch.isdigit() for ch in command):
            done = False
            while not done:
                word = line.pop()
                if word:
                    try:
                        node_id = int(word)
                    except ValueError:
                        pass
                    else:
                        self.display_node_by_id(node_id)
                else:
                    done = True
        elif command == "HELP":
            self.help(line)
        elif command == "EXECUTE":
            self.execute_cypher_from_file(line)
        elif command == "EXIT":
            raise StopIteration()
        elif command == "ADD":
            self.add_something(line)
        elif command == "CLEAR":
            if get_input("Are you sure you want to clear everything "
                         "from the database [y/N]? ").upper().startswith("Y"):
                print("Clearing all nodes and relationships")
                self.graph_db.clear()
            else:
                print("Clear aborted")
        elif command == "REMOVE":
            self.remove_something(line)
        elif command == "SHOW":
            self.show_something(line)
        elif command == "VERSION":
            self.show_neo4j_version(line)
        elif command == "LOOKUP":
            self.lookup(line)
        elif self.param_sets:
            self.execute_cypher(line.text, self.param_sets)
        else:
            self.execute_cypher(line.text, {})

    def help(self, line):
        sys.stdout.write(SHELL_HELP)

    def execute_cypher(self, query, params):
        if isinstance(params, list):
            for p in params:
                self.execute_cypher(query, p)
        else:
            if not isinstance(params, dict):
                params = {}
            try:
                record_set = neo4j.CypherQuery(self.graph_db, query).execute(**params)
            except CypherError as err:
                print("\x1b[31;1m{0}: {1}\x1b[0m".format(err.__class__.__name__, err))
                print("")
            else:
                writer = ResultWriter(sys.stdout)
                writer.write(self.format, record_set)

    def display_node_by_id(self, node_id):
        query = "START n=node({i}) RETURN n"
        params = {"i": node_id}
        try:
            record_set = neo4j.CypherQuery(self.graph_db, query).execute(**params)
        except CypherError as err:
            print("\x1b[31;1m{0}: {1}\x1b[0m".format(err.__class__.__name__, err))
            print("")
        else:
            self.display_node(record_set[0][0])

    def display_node(self, n):
        title = "Node {0}".format(n._id)
        print(title)
        print("=" * len(title))
        if self.graph_db.supports_node_labels:
            labels = n.get_labels()
            print("Labels: " + ", ".join(labels))
        print("Properties:")
        properties = n.get_cached_properties()
        max_key_len = max(len(key) for key in properties.keys())
        for key, value in sorted(properties.items()):
            print("  {0} : {1}".format(key.ljust(max_key_len),
                                     json.dumps(value)))
        print("Relationships:")
        for r in n.match():
            print("  {0}".format(r))
        print("")

    def lookup(self, line):
        command = line.pop()
        index_name = line.pop()
        if index_name:
            index = self.graph_db.get_index(neo4j.Node, index_name)
            if not index:
                print("\x1b[31;1mNode index {0} not found\x1b[0m".format(repr(index_name)))
                print("")
                return
        else:
            print("Usage: LOOKUP <index-name> <key> <value>")
            return
        key = line.pop()
        value = line.pop()
        nodes = index.get(key, value)
        if nodes:
            for n in nodes:
                self.display_node(n)
        else:
            print("No nodes found\n")

    def execute_cypher_from_file(self, line):
        command, file_name = self._pop_command_and_argument(line)
        if file_name:
            file_name = os.path.expanduser(file_name)
        else:
            print("Usage: EXECUTE <cypher-file>")
            return
        try:
            with codecs.open(file_name, encoding="utf-8") as f:
                query = f.read()
        except IOError as err:
            sys.stderr.write("{0}: {1}".format(err.__class__.__name__, err))
            sys.stderr.write("\n")
        else:
            if self.param_sets:
                self.execute_cypher(query, self.param_sets)
            else:
                self.execute_cypher(query, {})
            
    def show_neo4j_version(self, line):
        print("Neo4j " + self.graph_db.__metadata__["neo4j_version"])

    def add_something(self, line):
        command, subject = self._pop_command_and_argument(line)
        if not subject:
            print("Usage: ADD PARAMS <json-file>\n"
                  "       ADD PARAMETERS <json-file>")
            return
        if subject.upper() in ("PARAMS", "PARAMETERS"):
            self.add_parameters_from_file(line)
        else:
            sys.stderr.write("Bad command\n")

    def remove_something(self, line):
        command, subject = self._pop_command_and_argument(line)
        if not subject:
            print("Usage: REMOVE PARAMS\n"
                  "       REMOVE PARAMETERS")
            return
        if subject.upper() in ("PARAMS", "PARAMETERS"):
            self.remove_parameters(line)
        else:
            sys.stderr.write("Bad command\n")

    def show_something(self, line):
        command, subject = self._pop_command_and_argument(line)
        if not subject:
            print("Usage: SHOW PARAMS\n"
                  "       SHOW PARAMETERS")
            return
        if subject.upper() in ("PARAMS", "PARAMETERS"):
            self.show_parameters(line)
        else:
            sys.stderr.write("Bad command\n")

    def add_parameters_from_file(self, line):
        
        file_name = os.path.expanduser(line.pop())
        try:
            params = json.load(codecs.open(file_name, encoding="utf-8"))
        except IOError as err:
            sys.stderr.write("{0}: {1}".format(err.__class__.__name__, err))
            sys.stderr.write("\n")
        else:
            if isinstance(params, list):
                count = len(params)
                self.param_sets.extend(params)
            elif isinstance(params, dict):
                count = 1
                self.param_sets.append(params)
            else:
                count = 0
            if count == 1:
                print("1 parameter set added")
            else:
                print("{0} parameter sets added".format(count))

    def remove_parameters(self, line):
        self.param_sets = []

    def show_parameters(self, line):
        print(json.dumps(self.param_sets, sort_keys=True, indent=4))


if __name__ == "__main__":
    try:
        Tool().do(sys.argv)
        sys.exit(0)
    except Exception as err:
        sys.stderr.write(ustr(err))
        sys.stderr.write("\n")
        sys.exit(1)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


try:
    unicode
except NameError:
    # Python 3
    def ustr(s, encoding="utf-8"):
        if isinstance(s, str):
            return s
        try:
            return s.decode(encoding)
        except AttributeError:
            return str(s)
else:
    # Python 2
    def ustr(s, encoding="utf-8"):
        if isinstance(s, str):
            return s.decode(encoding)
        else:
            return unicode(s)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
The `calendar` module provides standardised date management functionality
based on a calendar subgraph::

    from py2neo import neo4j
    from py2neo.calendar import GregorianCalendar

    graph_db = neo4j.GraphDatabaseService()
    time_index = graph_db.get_or_create_index(neo4j.Node, "TIME")
    calendar = GregorianCalendar(time_index)

    graph_db.create(
        {"name": "Alice"},
        (0, "BORN", calendar.day(1800, 1, 1)),
        (0, "DIED", calendar.day(1900, 12, 31)),
    )

The root calendar node is held within a dedicated node index which needs
to be supplied to the calendar constructor.

All dates managed by the :py:class:`GregorianCalendar` class adhere to a
hierarchy such as::

    (CALENDAR)-[:YEAR]->(2000)-[:MONTH]->(12)-[:DAY]->(25)

"""


from __future__ import unicode_literals

from datetime import date as _date

from .neo4j import CypherQuery


class GregorianCalendar(object):

    class Date(object):

        def __init__(self, year, month=None, day=None):
            if year and month and day:
                self._date = _date(year, month, day)
                self._resolution = 3
                self.year = year
                self.month = month
                self.day = day
            elif year and month:
                self._date = _date(year, month, 1)
                self._resolution = 2
                self.year = year
                self.month = month
                self.day = None
            elif year:
                self._date = _date(year, 1, 1)
                self._resolution = 1
                self.year = year
                self.month = None
                self.day = None

        def __str__(self):
            if self.year and self.month and self.day:
                return "{y:04d}-{m:02d}-{d:02d}".format(
                    y=self.year, m=self.month, d=self.day
                )
            elif self.year and self.month:
                return "{y:04d}-{m:02d}".format(
                    y=self.year, m=self.month
                )
            elif self.year:
                return "{y:04d}".format(
                    y=self.year
                )

        def get_node(self, calendar):
            if self.year and self.month and self.day:
                return calendar.day(self.year, self.month, self.day)
            elif self.year and self.month:
                return calendar.month(self.year, self.month)
            elif self.year:
                return calendar.year(self.year)
            else:
                raise ValueError()

    class DateRange(object):

        def __init__(self, start_date=None, end_date=None):
            if start_date and end_date:
                self.start_date = GregorianCalendar.Date(*start_date)
                self.end_date = GregorianCalendar.Date(*end_date)
                # check both dates are compatible
                if self.start_date._resolution != self.end_date._resolution:
                    raise ValueError("Range start and end dates are specified "
                                     "at different resolutions")
                # ensure dates are correctly ordered
                if self.start_date._date > self.end_date._date:
                    self.start_date, self.end_date = self.end_date, self.start_date
            elif start_date:
                self.start_date = GregorianCalendar.Date(*start_date)
                self.end_date = None
            elif end_date:
                self.start_date = None
                self.end_date = GregorianCalendar.Date(*end_date)
            else:
                raise ValueError("Either start date or end date must "
                                 "be specified for a date range")

    def __init__(self, index):
        """ Create a new calendar instance pointed to by the index provided.
        """
        self._index = index
        self._graph_db = self._index.service_root.graph_db
        self._calendar = self._index.get_or_create("calendar", "Gregorian", {})

    def calendar(self):
        return self._calendar

    def day(self, year, month, day):
        """ Fetch the calendar node representing the day specified by `year`,
        `month` and `day`.
        """
        d = GregorianCalendar.Date(year, month, day)
        date_path = self._calendar.get_or_create_path(
            "YEAR",  {"year": d.year},
            "MONTH", {"year": d.year, "month": d.month},
            "DAY",   {"year": d.year, "month": d.month, "day": d.day},
        )
        return date_path.nodes[-1]

    def month(self, year, month):
        """ Fetch the calendar node representing the month specified by `year`
        and `month`.
        """
        d = GregorianCalendar.Date(year, month)
        date_path = self._calendar.get_or_create_path(
            "YEAR",  {"year": d.year},
            "MONTH", {"year": d.year, "month": d.month},
        )
        return date_path.nodes[-1]

    def year(self, year):
        """ Fetch the calendar node representing the year specified by `year`.
        """
        d = GregorianCalendar.Date(year)
        date_path = self._calendar.get_or_create_path(
            "YEAR",  {"year": d.year},
        )
        return date_path.nodes[-1]

    def date(self, date):
        return GregorianCalendar.Date(*date).get_node(self)

    def date_range(self, start_date=None, end_date=None):
        """ Fetch the calendar node representing the date range defined by
        `start_date` and `end_date`. If either are unspecified, this defines an
        open-ended range. Either `start_date` or `end_date` must be specified.
        """
        #                         (CAL)
        #                           |
        #                       [:RANGE]
        #                           |
        #                           v
        # (START)<-[:START_DATE]-(RANGE)-[:END_DATE]->(END)
        range_ = GregorianCalendar.DateRange(start_date, end_date)
        start, end = range_.start_date, range_.end_date
        if start and end:
            # if start and end are equal, return the day node instead
            if (start.year, start.month, start.day) == (end.year, end.month, end.day):
                return start.get_node(self)
            if (start.year, start.month) == (end.year, end.month):
                root = self.month(start.year, start.month)
            elif start.year == end.year:
                root = self.year(start.year)
            else:
                root = self._calendar
            query = """\
                START z=node({z}), s=node({s}), e=node({e})
                CREATE UNIQUE (s)<-[:START_DATE]-(r {r})-[:END_DATE]->(e),
                              (z)-[:DATE_RANGE]->(r {r})
                RETURN r
            """
            params = {
                "z": root._id,
                "s": start.get_node(self)._id,
                "e": end.get_node(self)._id,
                "r": {
                    "start_date": str(start),
                    "end_date": str(end),
                },
            }
        elif start:
            query = """\
                START z=node({z}), s=node({s})
                CREATE UNIQUE (s)<-[:START_DATE]-(r {r}),
                              (z)-[:DATE_RANGE]->(r {r})
                RETURN r
            """
            params = {
                "z": self._calendar._id,
                "s": start.get_node(self)._id,
                "r": {
                    "start_date": str(start),
                },
            }
        elif end:
            query = """\
                START z=node({z}), e=node({e})
                CREATE UNIQUE (r {r})-[:END_DATE]->(e),
                              (z)-[:DATE_RANGE]->(r {r})
                RETURN r
            """
            params = {
                "z": self._calendar._id,
                "e": end.get_node(self)._id,
                "r": {
                    "end_date": str(end),
                },
            }
        else:
            raise ValueError("Either start or end date must be supplied "
                             "for a date range")
        return CypherQuery(self._graph_db, query).execute_one(**params)

    def quarter(self, year, quarter):
        if quarter == 1:
            return self.date_range((year, 1, 1), (year, 3, 31))
        elif quarter == 2:
            return self.date_range((year, 4, 1), (year, 6, 30))
        elif quarter == 3:
            return self.date_range((year, 7, 1), (year, 9, 30))
        elif quarter == 4:
            return self.date_range((year, 10, 1), (year, 12, 31))
        else:
            raise ValueError("quarter must be in 1..4")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys

try:
    from setuptools import setup
    from setuptools.extension import Extension
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension

from py2neo import __author__, __email__, __license__, __package__, __version__


def do_setup(extensions=None):
    setup(
        name="py2neo-history",
        version=__version__,
        description="Python client library for the Neo4j REST server",
        long_description="Py2neo is a simple and pragmatic Python library that "
                         "provides access to the popular graph database Neo4j via "
                         "its RESTful web service interface. With no external "
                         "dependencies, installation is straightforward and "
                         "getting started with coding is easy. The library is "
                         "actively maintained on GitHub, regularly updated in the "
                         "Python Package Index and is built uniquely for Neo4j in "
                         "close association with its team and community.",
        author=__author__,
        author_email=__email__,
        url="http://nigelsmall.com/py2neo",
        scripts=[
            "scripts/neotool",
        ],
        packages=[
            "py2neo",
            "py2neo.packages",
            "py2neo.packages.httpstream",
            "py2neo.packages.jsonstream",
            "py2neo.packages.urimagic",
        ],
        install_requires=[
        ],
        license=__license__,
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 2.6",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.3",
            "Topic :: Database",
            "Topic :: Software Development",
        ],
        ext_modules=extensions,
        zip_safe=False,
    )


try:
    if sys.version_info >= (3,):
        do_setup([Extension("py2neo.packages.jsonstream.cjsonstream",
                            ["py2neo/packages/jsonstream/cjsonstream.c"])])
    else:
        do_setup([Extension("py2neo.packages.jsonstream.cjsonstream_2x",
                            ["py2neo/packages/jsonstream/cjsonstream_2x.c"])])
except:
    do_setup()
